/// <reference types="node" />
import JSZip from 'jszip';
import { ArchiveIndex, PartialTweetGDPR, PartialTweet, PartialTweetUser, InnerGDPRPersonalization, GPDRScreenNameHistory, GPDRProtectedHistory, InnerGDPRAgeInfo, GDPRMoment } from './TwitterTypes';
import DMArchive from './DMArchive';
import { EventTarget } from 'event-target-shim';
export declare type AcceptedZipSources = string | number[] | Uint8Array | ArrayBuffer | Blob | NodeJS.ReadableStream | JSZip | Archive;
export declare type ArchiveReadState = "idle" | "reading" | "indexing" | "tweet_read" | "user_read" | "dm_read" | "extended_read" | "ready";
/** Raw informations stored in GDPR, extracted for a simpler use.
 *
 * This includes list of followers, followings, favorites, mutes, blocks,
 * registered and subscribed lists, history of screen names, and Twitter moments.
 */
export interface ExtendedGDPRInfo {
    followers: Set<string>;
    followings: Set<string>;
    favorites: Set<string>;
    mutes: Set<string>;
    blocks: Set<string>;
    lists: {
        created: string[];
        member_of: string[];
        subscribed: string[];
    };
    personalization: InnerGDPRPersonalization;
    screen_name_history: GPDRScreenNameHistory[];
    protected_history: GPDRProtectedHistory[];
    age_info: InnerGDPRAgeInfo;
    moments: GDPRMoment[];
}
/** Return the `Date` object affiliated to **tweet**. */
export declare function dateFromTweet(tweet: PartialTweet): Date;
/**
 * Return true if **tweet** contains media(s).
 *
 * This includes photos, videos or animated GIF.
 */
export declare function isWithMedia(tweet: PartialTweet): boolean;
/**
 * Return true if **tweet** contains a video or one animated GIF.
 *
 * Twitter's GIF are mp4 encoded.
 */
export declare function isWithVideo(tweet: PartialTweet): boolean;
declare class Archive {
    protected _ready: Promise<void>;
    protected archive: JSZip;
    constructor(file: AcceptedZipSources);
    ready(): Promise<void>;
    dir(name: string): Archive;
    has(name: string): boolean;
    get(name: string, type?: "text" | "arraybuffer" | "blob", parse_auto?: boolean): Promise<any>;
    search(query: RegExp): JSZip.JSZipObject[];
    searchDir(query: RegExp): JSZip.JSZipObject[];
    read(file: JSZip.JSZipObject, type?: "text" | "arraybuffer" | "blob", parse_auto?: boolean): Promise<any>;
    ls(current_dir_only?: boolean): {
        [key: string]: JSZip.JSZipObject;
    };
    /**
     * Create a new instance of Archive from a file contained in this archive.
     *
     * @param name File name
     */
    fromFile(name: string): Promise<Archive>;
}
declare type TwitterArchiveEvents = {
    zipready: CustomEvent<void>;
    userinfosready: CustomEvent<void>;
    tweetsread: CustomEvent<void>;
    indexready: CustomEvent<void>;
    willreaddm: CustomEvent<void>;
    willreadextended: CustomEvent<void>;
    ready: CustomEvent<void>;
    error: CustomEvent<any>;
};
declare type TwitterArchiveOnEvents = {
    onzipready: CustomEvent<void>;
    onuserinfosready: CustomEvent<void>;
    ontweetsread: CustomEvent<void>;
    onindexready: CustomEvent<void>;
    onwillreaddm: CustomEvent<void>;
    onwillreadextended: CustomEvent<void>;
    onready: CustomEvent<void>;
    onerror: CustomEvent<any>;
};
/**
 * Represents a full TwitterArchive. Support GDPR and classic archive.
 *
 * Remember that tweets, in searchs in particular are **NOT** sorted.
 *
 * Direct messages, parsed if archive is a GDPR archive, stored in `.messages`,
 * are returned and sorted from the most older to the more recent.
 */
export declare class TwitterArchive extends EventTarget<TwitterArchiveEvents, TwitterArchiveOnEvents> {
    protected _ready: Promise<void>;
    protected archive: Archive;
    /** Current archive load state. */
    state: ArchiveReadState;
    /**
     * Access to extended infos. See `ExtendedGDPRInfo` interface.
     *
     * Defined only if `.is_gdpr === true`.
     * */
    extended_gdpr: ExtendedGDPRInfo;
    protected _index: ArchiveIndex;
    protected dms: DMArchive;
    protected dm_img_archive: Archive;
    protected dm_img_group_archive: Archive;
    protected _is_gdpr: boolean;
    protected user_cache: PartialTweetUser;
    /**
     * Build a new instance of TwitterArchive.
     *
     * @param file Accept any source that `JSZip` library accepts.
     * @param build_extended True if `.extended_gdpr` should be built (only if **file** is a GDPR archive.)
     */
    constructor(file: AcceptedZipSources, build_extended?: boolean);
    protected initGDPR(extended?: boolean): Promise<void>;
    protected initExtendedGDPR(): Promise<void>;
    protected convertToPartial(tweet: PartialTweetGDPR, pic_prof: string): PartialTweet;
    protected initClassic(): Promise<void>;
    protected isGDPRArchive(): boolean;
    /**
     * True if given tweet is coming from a GDPR archive.
     *
     * Tweets getted by available getters are NOT GDPR tweets, they've been converted !
     */
    isGDPRTweet(tweet: PartialTweetGDPR): true;
    isGDPRTweet(tweet: PartialTweet): false;
    /** Extract tweets from a specifc month. */
    month(month: string, year: string): PartialTweet[];
    /** Get tweets in a specific time interval. */
    between(since: Date, until: Date): PartialTweet[];
    /** Get a single tweet by ID. Returns `null` if tweet does not exists. */
    id(id_str: string): PartialTweet | null;
    /** Give the media url in direct message, obtain the Blob-bed image. This does **not** work on Node.js ! */
    dmImageFromUrl(url: string, is_group?: boolean): Promise<Blob>;
    /** Extract a direct message image from GDPR archive (exact filename required). This does **not** work on Node.js ! */
    dmImage(name: string, is_group?: boolean): Promise<Blob>;
    /** All tweets registered in this archive. */
    readonly all: PartialTweet[];
    /** Access to the DMArchive object. Will be undefined if `.is_gdpr === false`. */
    readonly messages: DMArchive;
    /** ID of the user who created this archive. */
    readonly owner: string;
    /**
     * Screen name (@) of the user who created this archive.
     * May be obsolete (user can change screen_name over time).
     */
    readonly owner_screen_name: string;
    /** Archive creation date. Not accurate in GDPR archive (will be the current date). */
    readonly generation_date: Date;
    /** Archive information and tweet index. */
    readonly index: ArchiveIndex;
    /** Number of tweets in this archive. */
    readonly length: number;
    /** True if archive is a GDPR archive. */
    readonly is_gdpr: boolean;
    /** Raw Archive object. Can be used to get specific files. */
    readonly raw: Archive;
    /** Resolved when archive read is over. */
    ready(): Promise<void>;
}
export {};
