"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const jszip_1 = __importDefault(require("jszip"));
const DMArchive_1 = __importDefault(require("./DMArchive"));
const event_target_shim_1 = require("event-target-shim");
const big_integer_1 = __importDefault(require("big-integer"));
const helpers_1 = require("./helpers");
/** Return the `Date` object affiliated to **tweet**. */
function dateFromTweet(tweet) {
    if (tweet.created_at_d) {
        return tweet.created_at_d;
    }
    return tweet.created_at_d = new Date(tweet.created_at);
}
exports.dateFromTweet = dateFromTweet;
/**
 * Return true if **tweet** contains media(s).
 *
 * This includes photos, videos or animated GIF.
 */
function isWithMedia(tweet) {
    return tweet.entities.media.length > 0;
}
exports.isWithMedia = isWithMedia;
/**
 * Return true if **tweet** contains a video or one animated GIF.
 *
 * Twitter's GIF are mp4 encoded.
 */
function isWithVideo(tweet) {
    if (tweet.extended_entities) {
        if (tweet.extended_entities.media) {
            return tweet.extended_entities.media.some(m => m.type !== "photo");
        }
    }
    return false;
}
exports.isWithVideo = isWithVideo;
class Archive {
    constructor(file) {
        this._ready = Promise.resolve();
        if (file instanceof Archive) {
            this._ready = Promise.resolve();
            this.archive = file.archive;
            return;
        }
        if (file instanceof jszip_1.default) {
            this._ready = Promise.resolve();
            this.archive = file;
            return;
        }
        this._ready = jszip_1.default.loadAsync(file)
            .then(data => {
            this.archive = data;
        });
    }
    ready() {
        return this._ready;
    }
    dir(name) {
        return new Archive(this.archive.folder(name));
    }
    has(name) {
        return this.search(new RegExp(`^${name}$`)).length > 0;
    }
    get(name, type = "text", parse_auto = true) {
        return this.read(this.archive.file(name), type, parse_auto);
    }
    search(query) {
        return this.archive.file(query);
    }
    searchDir(query) {
        return this.archive.folder(query);
    }
    read(file, type = "text", parse_auto = true) {
        const p = file.async(type);
        if (parse_auto) {
            return p.then(data => {
                if (typeof data === 'string') {
                    return JSON.parse(data.substr(data.indexOf('=') + 1).trimLeft());
                }
                else {
                    return data;
                }
            });
        }
        else {
            return p;
        }
    }
    ls(current_dir_only = true) {
        if (!current_dir_only) {
            return this.archive.files;
        }
        const l = this.archive.files;
        const files = {};
        for (const key in l) {
            if (!key.includes('/')) {
                files[key] = l[key];
            }
        }
        return files;
    }
    /**
     * Create a new instance of Archive from a file contained in this archive.
     *
     * @param name File name
     */
    async fromFile(name) {
        const f = await this.get(name, "arraybuffer", false);
        return new Archive(f);
    }
}
/**
 * Represents a full TwitterArchive. Support GDPR and classic archive.
 *
 * Remember that tweets, in searchs in particular are **NOT** sorted.
 *
 * Direct messages, parsed if archive is a GDPR archive, stored in `.messages`,
 * are returned and sorted from the most older to the more recent.
 */
class TwitterArchive extends event_target_shim_1.EventTarget {
    /**
     * Build a new instance of TwitterArchive.
     *
     * @param file Accept any source that `JSZip` library accepts.
     * @param build_extended True if `.extended_gdpr` should be built (only if **file** is a GDPR archive.)
     */
    constructor(file, build_extended = true) {
        super();
        this._ready = Promise.resolve();
        /** Current archive load state. */
        this.state = "idle";
        this._index = {
            info: {
                screen_name: "",
                full_name: "",
                location: "fr",
                bio: "",
                id: "",
                created_at: (new Date).toISOString()
            },
            archive: {
                created_at: (new Date).toISOString(),
                tweets: 0
            },
            years: {},
            by_id: {}
        };
        this._is_gdpr = false;
        this.state = "reading";
        this.archive = new Archive(file);
        this._ready = this.archive.ready()
            .then(() => {
            this.dispatchEvent({ type: 'zipready' });
            // Initialisation de l'archive Twitter
            if (this.isGDPRArchive()) {
                return this.initGDPR(build_extended);
            }
            else {
                return this.initClassic();
            }
        })
            .then(() => {
            this.dispatchEvent({ type: 'ready' });
        })
            .catch(e => {
            this.dispatchEvent({ type: 'error', detail: e });
            return Promise.reject(e);
        });
    }
    async initGDPR(extended = false) {
        this.state = "user_read";
        // Init informations
        const account_arr = await this.archive.get('account.js');
        const profile_arr = await this.archive.get('profile.js');
        const [account, profile] = [account_arr[0].account, profile_arr[0].profile];
        this.index.info = {
            screen_name: account.username,
            full_name: account.accountDisplayName,
            location: profile.description.location,
            bio: profile.description.bio,
            id: account.accountId,
            created_at: account.createdAt
        };
        this.dispatchEvent({ type: 'userinfosready' });
        this.state = "tweet_read";
        // Init tweet indexes
        const tweets = await this.archive.get('tweet.js');
        let i = 1;
        while (this.archive.has(`tweet-part${i}.js`)) {
            // Add every tweet in other files 
            // inside a "new" array in the initial array
            tweets.push(...(await this.archive.get(`tweet-part${i}.js`)));
            i++;
        }
        this.dispatchEvent({ type: 'tweetsread' });
        this.state = "indexing";
        // Build index
        for (let i = 0; i < tweets.length; i++) {
            const date = new Date(tweets[i].created_at);
            const month = String(date.getMonth() + 1);
            const year = String(date.getFullYear());
            // Creating month/year if not presents
            if (!(year in this.index.years)) {
                this.index.years[year] = {};
            }
            if (!(month in this.index.years[year])) {
                this.index.years[year][month] = {};
            }
            // Save tweet in index
            const converted = this.convertToPartial(tweets[i], profile.avatarMediaUrl);
            this.index.years[year][month][tweets[i].id_str] = converted;
            this.index.by_id[tweets[i].id_str] = converted;
        }
        this.dispatchEvent({ type: 'indexready' });
        // Register info
        this._index.archive.tweets = tweets.length;
        // Init DMs
        this.dms = new DMArchive_1.default(this.owner);
        this.state = "dm_read";
        this.dispatchEvent({ type: 'willreaddm' });
        const conversations = await this.archive.get('direct-message.js');
        this.dms.add(conversations);
        i = 1;
        while (this.archive.has(`direct-message-part${i}.js`)) {
            // Add every tweet in other files 
            // inside a "new" array in the initial array
            this.dms.add(await this.archive.get(`direct-message-part${i}.js`));
            i++;
        }
        if (this.archive.has('direct-message-group.js')) {
            this.dms.add(await this.archive.get('direct-message-group.js'));
        }
        // DMs should be ok
        // Test if archive contain DM images as zip
        if (this.archive.searchDir(new RegExp('direct_message_media')).length) {
            const folder = this.archive.dir('direct_message_media');
            const query = folder.search(/\.zip$/);
            if (query.length) {
                // console.log("Creating archive from archive (single)");
                this.dm_img_archive = await this.archive.fromFile(query[0].name);
            }
        }
        if (this.archive.searchDir(new RegExp('direct_message_group_media')).length) {
            const folder = this.archive.dir('direct_message_group_media');
            const query = folder.search(/\.zip$/);
            if (query.length) {
                // console.log("Creating archive from archive (group)");
                this.dm_img_group_archive = await this.archive.fromFile(query[0].name);
            }
        }
        this.state = "extended_read";
        this.dispatchEvent({ type: 'willreadextended' });
        if (extended) {
            await this.initExtendedGDPR();
        }
        this.state = "ready";
    }
    async initExtendedGDPR() {
        // Followings
        const followings = new Set();
        try {
            const f_following = await this.archive.get('following.js');
            for (const f of f_following) {
                followings.add(f.following.accountId);
            }
        }
        catch (e) { }
        // Followers
        const followers = new Set();
        try {
            const f_follower = await this.archive.get('follower.js');
            for (const f of f_follower) {
                followers.add(f.follower.accountId);
            }
        }
        catch (e) { }
        // Favorites
        const favorites = new Set();
        try {
            const f_fav = await this.archive.get('like.js');
            for (const f of f_fav) {
                favorites.add(f.like.tweetId);
            }
        }
        catch (e) { }
        // Mutes
        const mutes = new Set();
        try {
            const f_mutes = await this.archive.get('mute.js');
            for (const f of f_mutes) {
                mutes.add(f.muting.accountId);
            }
        }
        catch (e) { }
        // Blocks
        const blocks = new Set();
        try {
            const f_block = await this.archive.get('block.js');
            for (const f of f_block) {
                blocks.add(f.blocking.accountId);
            }
        }
        catch (e) { }
        // Lists
        const lists = {
            created: [],
            member_of: [],
            subscribed: []
        };
        try {
            lists.created = (await this.archive.get('lists-created.js'))[0].userListInfo.urls;
            lists.member_of = (await this.archive.get('lists-member.js'))[0].userListInfo.urls;
            lists.subscribed = (await this.archive.get('lists-subscribed.js'))[0].userListInfo.urls;
        }
        catch (e) { }
        // Personalization
        let personalization = {
            demographics: undefined,
            locationHistory: [],
            interests: undefined
        };
        try {
            personalization = (await this.archive.get('personalization.js'))[0].p13nData;
        }
        catch (e) { }
        let age_info;
        try {
            age_info = (await this.archive.get('ageinfo.js'))[0].ageMeta;
        }
        catch (e) { }
        // SN history
        const screen_name_history = [];
        try {
            const f_history = await this.archive.get('screen-name-change.js');
            for (const e of f_history) {
                screen_name_history.push(e.screenNameChange);
            }
        }
        catch (e) { }
        // Protected history
        const protected_history = [];
        try {
            const f_phistory = await this.archive.get('protected-history.js');
            for (const e of f_phistory) {
                protected_history.push(e.protectedHistory);
            }
        }
        catch (e) { }
        // Moments
        const moments = (await this.archive.get('moment.js')).map(e => e.moment);
        this.extended_gdpr = {
            moments,
            protected_history,
            screen_name_history,
            age_info,
            personalization,
            lists,
            favorites,
            followers,
            followings,
            mutes,
            blocks
        };
    }
    convertToPartial(tweet, pic_prof) {
        if (!this.user_cache) {
            this.user_cache = {
                protected: false,
                id_str: this._index.info.id,
                name: this._index.info.full_name,
                screen_name: this._index.info.screen_name,
                profile_image_url_https: pic_prof
            };
        }
        tweet.user = this.user_cache;
        tweet.text = tweet.full_text;
        // Gérer le cas des retweets
        const rt_data = /^RT @(.+): (.+)/.exec(tweet.full_text);
        if (rt_data && rt_data.length && !("retweeted_status" in tweet)) {
            const [, arobase, text] = rt_data;
            const rt = Object.assign({}, tweet);
            rt.user = Object.assign({}, rt.user);
            rt.text = text;
            rt.user.screen_name = arobase;
            rt.user.name = arobase;
            // @ts-ignore
            rt.retweeted = true;
            // Recherche si un ID est disponible par exemple dans les medias (sinon tant pis)
            if (rt.extended_entities && rt.extended_entities.media) {
                if (rt.extended_entities.media.length) {
                    rt.user.id_str = rt.extended_entities.media[0].source_user_id_str;
                }
            }
            tweet.retweeted_status = rt;
        }
        return tweet;
    }
    async initClassic() {
        this.state = "user_read";
        const js_dir = this.archive.dir('data').dir('js');
        const index = await js_dir.get('tweet_index.js');
        const payload = await js_dir.get('payload_details.js');
        const user = await js_dir.get('user_details.js');
        this.index.info = user;
        this.index.archive = payload;
        this.dispatchEvent({ type: 'userinfosready' });
        const files_to_read = index.map(e => e.file_name);
        let tweets = [];
        this.state = "tweet_read";
        const tweet_file_promises = [];
        for (const file of files_to_read) {
            tweet_file_promises.push(this.archive.get(file).then((t) => {
                tweets.push(...t);
            }));
        }
        await Promise.all(tweet_file_promises);
        // Tri les tweets par ID (le plus récent, plus grand en premier)
        if (helpers_1.supportsBigInt()) {
            tweets = tweets.sort((a, b) => Number(BigInt(b.id_str) - BigInt(a.id_str)));
        }
        else {
            tweets = tweets.sort((a, b) => (big_integer_1.default(b.id_str).minus(big_integer_1.default(a.id_str))).toJSNumber());
        }
        this.dispatchEvent({ type: 'tweetsread' });
        this.state = "indexing";
        // Build index (read tweets)
        for (let i = 0; i < tweets.length; i++) {
            const date = new Date(tweets[i].created_at);
            const month = String(date.getMonth() + 1);
            const year = String(date.getFullYear());
            // Creating month/year if not presents
            if (!(year in this.index.years)) {
                this.index.years[year] = {};
            }
            if (!(month in this.index.years[year])) {
                this.index.years[year][month] = {};
            }
            // Save tweet in index
            this.index.years[year][month][tweets[i].id_str] = tweets[i];
            this.index.by_id[tweets[i].id_str] = tweets[i];
        }
        // Setting right tweet number
        this.index.archive.tweets = Object.keys(this.index.by_id).length;
        this.dispatchEvent({ type: 'indexready' });
        this.state = "ready";
    }
    isGDPRArchive() {
        return this._is_gdpr = this.archive.search(/^tweets\.csv$/).length === 0;
    }
    isGDPRTweet(tweet) {
        return 'retweet_count' in tweet;
    }
    /** Extract tweets from a specifc month. */
    month(month, year) {
        if (year in this.index.years) {
            if (month in this.index.years[year]) {
                return Object.values(this.index.years[year][month]);
            }
        }
        return [];
    }
    /** Get tweets in a specific time interval. */
    between(since, until) {
        if (since.getTime() > until.getTime()) {
            throw new Error("Since can't be superior to until");
        }
        const firsts = [String(since.getMonth() + 1), String(since.getFullYear())];
        const ends = [String(until.getMonth() + 1), String(until.getFullYear())];
        const remaining_months = [];
        remaining_months.push(firsts);
        let tmp_date = new Date(since.getFullYear(), since.getMonth());
        tmp_date.setMonth(tmp_date.getMonth() + 1);
        remaining_months.push(ends);
        // Calculating months that are between the two dates
        while (true) {
            if (tmp_date.getFullYear() > until.getFullYear()) {
                break;
            }
            if (tmp_date.getFullYear() === until.getFullYear()) {
                if (tmp_date.getMonth() >= until.getMonth()) {
                    break;
                }
            }
            remaining_months.push([String(tmp_date.getMonth() + 1), String(tmp_date.getFullYear())]);
            tmp_date.setMonth(tmp_date.getMonth() + 1);
        }
        const tweets = [];
        const [end_day, end_month, end_year] = [until.getDate(), until.getMonth(), until.getFullYear()];
        // Finding tweets that are same or after since date or before until date
        for (const [month, year] of remaining_months) {
            const m = this.month(month, year);
            for (const t of m) {
                const d = dateFromTweet(t);
                if (d.getTime() >= since.getTime()) {
                    if (d.getFullYear() < end_year || d.getMonth() < end_month || d.getDate() <= end_day) {
                        tweets.push(t);
                    }
                }
            }
        }
        return tweets;
    }
    /** Get a single tweet by ID. Returns `null` if tweet does not exists. */
    id(id_str) {
        if (id_str in this.index.by_id) {
            return this.index.by_id[id_str];
        }
        return null;
    }
    /** Give the media url in direct message, obtain the Blob-bed image. This does **not** work on Node.js ! */
    dmImageFromUrl(url, is_group = false) {
        const [, , , , id, , image] = url.split('/');
        if (id && image) {
            return this.dmImage(id + "-" + image, is_group);
        }
        return Promise.reject();
    }
    /** Extract a direct message image from GDPR archive (exact filename required). This does **not** work on Node.js ! */
    dmImage(name, is_group = false) {
        if (!this.is_gdpr) {
            return Promise.reject("Archive not supported");
        }
        if (this.dm_img_archive || this.dm_img_group_archive) {
            // Les dm sont dans un ZIP
            let zip;
            if (is_group) {
                zip = this.dm_img_group_archive;
            }
            else {
                zip = this.dm_img_archive;
            }
            if (!zip) {
                return Promise.reject("Image archive is not loaded, impossible to get media.");
            }
            const results = zip.search(new RegExp(name + "(\.?.*)$"));
            if (results.length) {
                return zip.read(results[0], "blob");
            }
            return Promise.reject("File not found");
        }
        else {
            const directory = this.archive.dir(is_group ? "direct_message_group_media" : "direct_message_media");
            const results = directory.search(new RegExp(name + "(\.?.*)$"));
            if (results.length) {
                return this.archive.read(results[0], "blob");
            }
            return Promise.reject("File not found");
        }
    }
    /** All tweets registered in this archive. */
    get all() {
        return Object.values(this.index.by_id);
    }
    /** Access to the DMArchive object. Will be undefined if `.is_gdpr === false`. */
    get messages() {
        return this.dms;
    }
    /** ID of the user who created this archive. */
    get owner() {
        return this._index.info.id;
    }
    /**
     * Screen name (@) of the user who created this archive.
     * May be obsolete (user can change screen_name over time).
     */
    get owner_screen_name() {
        return this._index.info.screen_name;
    }
    /** Archive creation date. Not accurate in GDPR archive (will be the current date). */
    get generation_date() {
        return new Date(this._index.archive.created_at);
    }
    /** Archive information and tweet index. */
    get index() {
        return this._index;
    }
    /** Number of tweets in this archive. */
    get length() {
        return this.index.archive.tweets;
    }
    /** True if archive is a GDPR archive. */
    get is_gdpr() {
        return this._is_gdpr;
    }
    /** Raw Archive object. Can be used to get specific files. */
    get raw() {
        return this.archive;
    }
    /** Resolved when archive read is over. */
    ready() {
        return this._ready;
    }
}
exports.TwitterArchive = TwitterArchive;
// Define onevents
event_target_shim_1.defineEventAttribute(TwitterArchive.prototype, "zipready");
event_target_shim_1.defineEventAttribute(TwitterArchive.prototype, "userinfosready");
event_target_shim_1.defineEventAttribute(TwitterArchive.prototype, "tweetsread");
event_target_shim_1.defineEventAttribute(TwitterArchive.prototype, "indexready");
event_target_shim_1.defineEventAttribute(TwitterArchive.prototype, "willreaddm");
event_target_shim_1.defineEventAttribute(TwitterArchive.prototype, "willreadextended");
event_target_shim_1.defineEventAttribute(TwitterArchive.prototype, "ready");
event_target_shim_1.defineEventAttribute(TwitterArchive.prototype, "error");
